#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <memory.h>

#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/epoll.h>
#include <netinet/tcp.h>

#include <string>
#include <functional>
#include <thread>

#include <a8/a8.h>
#include <a8/udplistener.h>

#if 0
#include <f8/udplog.h>
#endif

namespace a8
{

    struct UdpListenerImpl
    {
        a8::UdpListener* master = nullptr;

        int listen_socket = -1;
        std::thread* worker_thread = nullptr;

        volatile bool shutdown = false;

        virtual ~UdpListenerImpl()
        {
            if (IsActive()) {
                SetActive(false);
            }
        }

        bool IsActive()
        {
            return listen_socket != -1;
        }

        void SetActive(bool active)
        {
            if (active) {
                if (IsActive()) {
                    return;
                }
                if (!worker_thread) {
                    shutdown = false;
                    worker_thread = new std::thread(&a8::UdpListenerImpl::WorkerThreadProc, this);
                }
            } else {
                if (IsActive()) {
                    ActiveStop();
                }
            }
        }

        bool ActiveStart()
        {
            listen_socket = ::socket(AF_INET, SOCK_DGRAM, 0);
            if (listen_socket == -1){
                if (master->on_error) {
                    master->on_error(errno);
                }
                return false;
            }
            // TIME_WAIT - argh
            int on = 1;
            if (::setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, (const char*)&on, sizeof(on)) == -1 ){
                if (master->on_error){
                    master->on_error(errno);
                }
                ::close(listen_socket);
                listen_socket = a8::INVALID_SOCKET;
                return false;
            }
            sockaddr_in sa;
            memset(&sa, 0, sizeof(sa));
            sa.sin_family = AF_INET;
            sa.sin_addr.s_addr = inet_addr(master->bind_address.c_str());
            sa.sin_port = htons(master->bind_port);
            if (::bind(listen_socket, (sockaddr*)&sa, sizeof(sa)) < 0) {
                if (master->on_error) {
                    master->on_error(errno);
                }
                ::close(listen_socket);
                listen_socket = -1;
                return false;
            }
            return true;
        }

        void ActiveStop()
        {
            if (listen_socket != -1) {
                ::shutdown(listen_socket, SHUT_RD);
                ::close(listen_socket);
                listen_socket = -1;
            }
            if (worker_thread) {
                shutdown = true;
                worker_thread->join();
                delete worker_thread;
                worker_thread = NULL;
            }
        }

        void WorkerThreadProc()
        {
            if(!ActiveStart())
                return;
            struct sockaddr_in from_addr;
            socklen_t addr_len = sizeof(sockaddr_in);
            char recv_buf[1024 * 64];
            int buflen = sizeof(recv_buf);
            while (!shutdown) {
                int ret = ::recvfrom(listen_socket,
                                     recv_buf,
                                     buflen,
                                     0,
                                     (struct sockaddr*)&from_addr,
                                     (socklen_t*)&addr_len);
                if(ret > 0) {
                    #if 0
                    f8::UdpLog::Instance()->Debug("recvfrom host:%s port:%d buflen:%d",
                                                  {
                                                   inet_ntoa(from_addr.sin_addr),
                                                   from_addr.sin_port,
                                                   ret
                                                  });
                    #endif
                    a8::UdpPacket* pkt = new a8::UdpPacket();
                    pkt->buf = (char*)malloc(ret + 1);
                    memmove((void*)pkt->buf, recv_buf, ret);
                    pkt->buf_len = ret;
                    pkt->remote_addr = from_addr;
                    master->on_recv_packet(pkt);
                } else {
                    #if 0
                    f8::UdpLog::Instance()->Debug("recvfrom error ret:%d",
                                                  {
                                                   ret
                                                  });
                    #endif
                }
            }
        }

    };

    UdpListener::UdpListener()
    {
        impl_ = std::make_shared<a8::UdpListenerImpl>();
        impl_->master = this;
    }

    UdpListener::~UdpListener()
    {
        impl_ = nullptr;
    }

    void UdpListener::Open()
    {
        if (!IsActive()) {
            impl_->SetActive(true);
        }
    }

    void UdpListener::Close()
    {
        if (IsActive()) {
            impl_->SetActive(false);
        }
    }

    bool UdpListener::IsActive()
    {
        return impl_->IsActive();
    }

    void UdpListener::SendUdpPacket(UdpPacket* pkt)
    {
        ::sendto(impl_->listen_socket,
                 pkt->buf,
                 pkt->buf_len,
                 0,
                 (struct sockaddr*)&pkt->remote_addr, sizeof(pkt->remote_addr));

    }

}
